/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: cal_table.c,v 1.5 2002/11/07 12:50:39 jaffe Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: cal_table.c,v 1.5 2002/11/07 12:50:39 jaffe Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*----------------------------------------------------------------------- */
/* Calibration table for optical/IR interferometer                        */
/*----------------------------------------------------------------------- */
#include "cal_table.h"

/* internal prototypes */
/* get data from table_def specific to cal_table table */
void get_cal_table_info (cal_table_def *table);

/* file global variables for table type*/
static char *tabletype = "CALIBRATION";


/* cal_table_row functions */
cal_table_row* make_cal_table_row(cal_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an cal_table_row structure                                      */
/*   Inputs:                                                              */
/*      table     pointer to cal_table table structure                    */
/*   Returns:                                                             */
/*      An cal_table_row structure pointer ; NULL indicates failure       */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  int mpoly;
  cal_table_row *out;

  /* create basic structure */
  out = (cal_table_row*)malloc(sizeof(cal_table_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(cal_table_row));

  /* polynomial */
  mpoly = table->npoly;
  if (mpoly<1) mpoly = 1; /* at least one even if not used */
  out->amp_poly = (float*)malloc(mpoly*sizeof(float));
  
  /* Make generic row structure */
  out->raw = make_table_row_from_def(table->def);

  return out; /* done */
} /* end  make_cal_table_row */

void kill_cal_table_row (cal_table_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an cal_table_row structure                                    */
/*   Inputs:                                                              */
/*      row    pointer to the cal_table_row structure.                    */
/*----------------------------------------------------------------------- */
{
  if (row==NULL) return; /* anybody home */
  if (row->raw!=NULL) free (row->raw); /* generic row structure */
  if (row->amp_poly!=NULL) free(row->amp_poly);  /* polynomial */
  free (row);                          /* basic structure */
} /* end  kill_cal_table_row */


/* cal_table functions */
cal_table_def* 
make_cal_table_def(fitsfile *fptr, char *date_obs, int npoly, int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new cal_table in an open FITS file                            */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      date_obs  Date (yyyy-mm-dd) observations began                    */
/*      npoly     number of terms in teh amplitude polynomial             */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      A cal_table_def pointer; NULL indicates failure                   */
/*----------------------------------------------------------------------- */
{
  cal_table_def *out = NULL;
  int                    version = 1; /* table version number */
  int                    ikey, icol, i;
  long                   nx, ny, nregion;

  if (*ierr>0) return NULL;    /* existing error condition? */

  if (fptr==NULL)     return NULL; /* sanity check */

  /* create output object */
  out = (cal_table_def*)malloc(sizeof(cal_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(cal_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(CALIBRATION_TABLE, 0, 0);
  if (out->def==NULL) { /* check for failure */
    kill_cal_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating IMAGING_DATA_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* number of polynomial terms */
  out->npoly = npoly;
  icol = table_find_col (out->def, "AMP_POLY")-1;
  if (icol>0) out->def->colcount[icol] = npoly;

  /* add keyword values */
  ikey = table_find_key(out->def, "REVISION");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&version, ierr);
  ikey = table_find_key(out->def, "NPOLY");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&npoly, ierr);
  ikey = table_find_key(out->def, "DATE-OBS");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)date_obs, ierr);

  /* get info from table descriptor */
  get_cal_table_info (out);

  /* open calibration table */
  make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_cal_table */

cal_table_def* open_cal_table(fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing cal_table in the active FITS file                   */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An cal_table_def pointer; NULL indicates failure                  */
/*----------------------------------------------------------------------- */
{
  cal_table_def *out = NULL;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (cal_table_def*)malloc(sizeof(cal_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(cal_table_def)); /* zero fill */

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_cal_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_cal_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get info from table descriptor */
  get_cal_table_info (out);

  return out; /* done */
} /* end open_cal_table */

int read_cal_table(fitsfile *fptr, cal_table_def* table, 
			    cal_table_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an cal_table                                  */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  int i;
  table_row *trow;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  read_cal_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  trow = row->raw;
  row->time        = trow->data[table->time_col]->tdata.tdouble[0];
  row->timei       = trow->data[table->interv_col]->tdata.tfloat[0];
  row->gain[0]     = trow->data[table->gain_col]->tdata.tfloat[0];
  row->gain[1]     = trow->data[table->gain_col]->tdata.tfloat[1];
  row->delay       = trow->data[table->delay_col]->tdata.tfloat[0];
  row->rate        = trow->data[table->rate_col]->tdata.tfloat[0];
  row->weight      = trow->data[table->weight_col]->tdata.tfloat[0];
  row->source      = trow->data[table->source_col]->tdata.tshort[0];
  row->tele        = trow->data[table->tele_col]->tdata.tshort[0];
  row->ref         = trow->data[table->ref_col]->tdata.tshort[0];
  for (i=0; i<table->npoly; i++)
    row->amp_poly[i] = trow->data[table->poly_col]->tdata.tfloat[i];

  return 0; /* OK - done */
} /* end read_cal_table */

int write_cal_table(fitsfile *fptr, cal_table_def* table, 
			     cal_table_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an cal_table                                  */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  trow = row->raw;
  trow->data[table->time_col]->tdata.tdouble[0]  = row->time;
  trow->data[table->interv_col]->tdata.tfloat[0] = row->timei;
  trow->data[table->gain_col]->tdata.tfloat[0]   = row->gain[0];
  trow->data[table->gain_col]->tdata.tfloat[1]   = row->gain[1];
  trow->data[table->delay_col]->tdata.tfloat[0]  = row->delay;
  trow->data[table->rate_col]->tdata.tfloat[0]   = row->rate;
  trow->data[table->weight_col]->tdata.tfloat[0] = row->weight;
  trow->data[table->source_col]->tdata.tshort[0] = row->source;
  trow->data[table->tele_col]->tdata.tshort[0]   = row->tele;
  trow->data[table->ref_col]->tdata.tshort[0]    = row->ref;
  for (i=0; i<table->npoly; i++)
    trow->data[table->poly_col]->tdata.tfloat[i] = row->amp_poly[i];

  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_cal_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_cal_table */

int close_cal_table(fitsfile *fptr, cal_table_def* table, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open cal_table, associated structures are destroyed*/
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_cal_table_def(table);

  return 0; /* OK - done */
} /* end close_cal_table */

void kill_cal_table_def (cal_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an cal_table structure                               */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  kill_table_def(table->def);

  /* free keyword arrays */
  if (table->date_obs!=NULL) free(table->date_obs);

  /* free basic structure */
  free (table);
} /* end kill_cal_table_def */

/*************************private functions ******************************/

void get_cal_table_info (cal_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to cal_table table                  */
/*   Uses detector member which must already be attached                  */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  int ikey, ierr=0;
  char *date_obs="unknown";
  if (table==NULL) return;             /* sanity check */

  /* keyword values */
  ikey = table_find_key(table->def, "DATE-OBS");
  if (ikey>0) date_obs = (char*)table->def->keydata[ikey-1];
  table->date_obs = (char*)malloc(strlen(date_obs)+1);
  strcpy (table->date_obs, date_obs);
  ikey = table_find_key(table->def, "NPOLY");
  table->npoly = 0;
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->npoly, &ierr);

  /* lookup column numbers (0-rel) */
  table->time_col    = table_find_col (table->def, "TIME")-1;
  table->interv_col  = table_find_col (table->def, "TIME_INTE")-1;
  table->source_col  = table_find_col (table->def, "SOURCE_ID")-1;
  table->tele_col    = table_find_col (table->def, "TELESCOPE")-1;
  table->poly_col    = table_find_col (table->def, "AMP_POLY")-1;
  table->gain_col    = table_find_col (table->def, "GAIN")-1;
  table->delay_col   = table_find_col (table->def, "DELAY")-1;
  table->rate_col    = table_find_col (table->def, "RATE")-1;
  table->weight_col  = table_find_col (table->def, "WEIGHT")-1;
  table->ref_col     = table_find_col (table->def, "REFERENCE")-1;

} /* end  get_cal_table_info */
/* */
